/* Name: main.c
 * Author: Michael Ciuffo
 * Published under Creative Commons Attribution-NonCommercial 3.0 Unported license
 */

#include <avr/io.h>
#include <avr/pgmspace.h>
#include <util/delay.h>
#include <avr/interrupt.h>  
#include <avr/sleep.h>

const uint8_t flicker[100] PROGMEM = {0x3e, 0x17, 0xe9, 0x31, 0x29, 0xa1, 0x4f, 0x7d, 0x66, 0x15, 0x2e, 0x25, 0x17, 0x33, 0x31, 0x2, 0x2d, 0x68, 0xa5, 0x64, 0x2, 0x55, 0x9f, 0xfd, 0xa7, 0xa0, 0x67, 0x94, 0x9a, 0xc6, 0x7c, 0xd0, 0xe9, 0x81, 0x3f, 0x8d, 0xc8, 0x87, 0x20, 0xb5, 0xf0, 0xdc, 0xd9, 0x10, 0xc2, 0x5d, 0x15, 0x1a, 0x37, 0x9c, 0xf1, 0xa4, 0x97, 0xb9, 0xf3, 0xad, 0x26, 0xe5, 0x46, 0xf2, 0x2e, 0xcd, 0xf4, 0xec, 0xf5, 0xbd, 0x7d, 0x27, 0xb6, 0xec, 0xc8, 0xa, 0x87, 0x5e, 0xbb, 0xb0, 0x1d, 0x48, 0x49, 0x6a, 0x60, 0xce, 0xd6, 0x3d, 0x1d, 0x2a, 0x28, 0x64, 0x71, 0xa5, 0x26, 0xcd, 0xcb, 0x4b, 0x1a, 0x24, 0xcc, 0xe9, 0x68,0x77};
uint32_t touchingthreshold;	//discharge time under which the micro registers a touch
uint32_t letgothreshold;	//discharge time over which the micro registers a letgo

const uint32_t defaulttouchthreshold = 1000;	//these were placed so that you can change the discharge time
const uint32_t defaultletgothreshold = 2000;	//on the fly for an adjustable threshold (not implemented)

const uint32_t maxwaittime = 10000; //maximum time you will wait for someone to let go

uint8_t randomindex; //which random number is currently being used
uint8_t Acandles;	//which candles on Port A should be lit (note that 0 is "lit")
uint8_t Bcandles;   //which candles on Port B should be lit (note that 0 is "lit")
uint8_t day;		//current day.
uint8_t sleep;		//should the unit be sleeping?
uint8_t shutdown;	//if true, device will shut down as soon as possible.

const uint32_t burndowntime = 90000; //how long before it sleeps.  100 is around 12 seconds.


//Returns the next random number in the series.
uint8_t nextrandom()
{
	if (randomindex<100)
		randomindex++;
	else
		randomindex = 0;
	return pgm_read_byte(&(flicker[randomindex]));
}

//advance day counter by one
void nextday(void)
{
	if (day<8)
		day++;
	else
	{
		shutdown = 1;
	}
	changeday(day);
}

//go through all the steps of charging up the touchsense pad.
//This is "slow" because running this routine necessitates 
//doing all of the steps of pushing and popping stuff off the stack
//which takes time.  This routine won't work if you're actually trying to measure
//the discharge time, but it's good for casting off the capsense to wait for the next hit.
void slowcharge(void)
{
	DDRB = 0xFF;
	PORTB = PORTB | 0xFD; //charge capsense (don't charge pin 1)
	DDRB = 0x02;			//change capsense to input
	PORTB = PORTB & 0x02; //drop capsense pin
	
}


//routine loops until you let go of the pad.
void letgo(uint16_t time)
{	
	uint16_t count = 0;
	uint8_t loop = 1;
	uint16_t attemptcount = 0;
	while (loop)
	{
		count = 0;
		DDRB = 0xFF;

		PORTB = PORTB | 0xFD; //charge capsense (don't charge pin 1)
		DDRB = 0x02;			//change capsense to input
		PORTB = PORTB & 0x02; //drop capsense pin
		while (PINB & 0X04)//wait for it to fall
		{
			count++;
			attemptcount++;
		}
		if (count > (letgothreshold)) //did it fall slow enough?
			loop = 0;
	}
}

//change day to day in argument.
void changeday(uint8_t day)
{
	switch (day) {
		case 1:
			Acandles = 0xAF;
			Bcandles = 0xFF;
			break;
		case 2:
			Acandles = 0x2F;
			Bcandles = 0xFF;
			break;
		case 3:
			Acandles = 0x2F;
			Bcandles = 0xFD;
			break;
		case 4:
			Acandles = 0x0F;
			Bcandles = 0xFD;
			break;
		case 5:
			Acandles = 0x07;
			Bcandles = 0xFD;
			break;
		case 6:
			Acandles = 0x03;
			Bcandles = 0xFD;
			break;
		case 7:
			Acandles = 0x01;
			Bcandles = 0xFD;
			break;
		case 8:
			Acandles = 0x00;
			Bcandles = 0xFD;
			break;
		default:
			Acandles = 0xff;
			Bcandles = 0xFD;
			break;
	}

}

//returns true if the user is touching the pad.
uint8_t touching(void)
{
	uint16_t count=0;
	DDRB = 0xFF;
	PORTB = PORTB | 0xFD; //charge capsense (don't charge pin 1)
	DDRB = 0x02;			//change capsense to input
	PORTB = PORTB & 0x02; //drop capsense pin
	while ((PINB & 0X04)&& (count < maxwaittime)) //wait for it to fall
		count++;
	if (count < touchingthreshold)
		return 1;
	else
		return 0;
}
		

//interrupt handler.  Is called every time user touches pad (including wakeups)
//doesn't do anything if the unit is supposed to be asleep though.
ISR (INT0_vect)
{ 
	if (!sleep)
	{
		cli();
		PORTA = Acandles;	//to provide good feedback, turn all relevant candles on full
		PORTB = Bcandles;	//without this, you catch them mid-PWM and it's confusing
							//for the user
		_delay_ms(1);		//a nifty delay.  Helps make sure the finger is all the way down before
							//measurement.
		uint16_t count = 0;
		
		//this next section can not be done in slowcharge because too much time is spent
		//exiting that routine and it messes up "count".
		
		DDRB = 0xFF;
		PORTB = PORTB | 0xFD; //charge capsense (don't charge pin 1)
		DDRB = 0x02;			//change capsense to input
		PORTB = PORTB & 0x02; //drop capsense pin
		while ((PINB & 0X04)&& (count < maxwaittime)) //wait for it to fall, but if it doesn't don't keep waiting.
			count++;
		if (count < touchingthreshold) //a reasonable threshold that I figured out by guess and check
		{	
			nextday();    //user pressed the button, advance a day.
			display();	  //display the current day.  It's nice to have quick visual feedback rather
						  //than waiting for the main routine loop.
			
			letgo(count);	  //wait for the user to let go.
		}
		
		slowcharge();	//charge cap for next time.
		GIFR = (1<<INTF0);	//clear interrupt flag
		sei();			//enable interrupts.
	}
}


//display current day
void display(void)
{
	PORTA = Acandles;
	PORTB = Bcandles;
}


int main(void)
{
	DDRB =  0xFF;	//All of port A is used, only pin 1 of port B is.
	PORTB = 0X00;	//note that 1 is off and 0 is on
	DDRA = 0XFF;
	PORTA = 0X00;

	day = 1;
	changeday(day);
	slowcharge();
	shutdown = 0;
	letgothreshold = defaultletgothreshold;
	touchingthreshold = defaulttouchthreshold;
	MCUCR = 0x10; //POWER-DOWN sleep mode. //low level interrupt.
	
	GIMSK = 0x40; //INT0 enabled
	sei();

    for(;;)
	{
		//This is the main loop
		//when it's done its thing n times, sleep.
		uint32_t time = 0;  
		while((time < burndowntime) && !shutdown)
		{
			time++;
			int i;
			uint8_t j;
			int currentrand;
			uint8_t flickcounter = 0x0f;
			uint8_t outputbufferA = 0xff;
			uint8_t outputbufferB = 0x02;
			uint8_t outputvalues[10];
			for (i = 0; i<10; i+=2)		//grab 10 nibbles from the random table
			{
				currentrand = nextrandom();
				outputvalues[i] = (0x0f&currentrand);
				outputvalues[i+1] = ((0xf0&currentrand)>>4);
			}
			for (i = 0; i<30; i++)
			{
				while (flickcounter > 0) //step through the numbers 0 through 7
				{
					outputbufferA = 0x00;
					outputbufferB = 0x00;
					for (j=0; j<8; j++)
					{
						if (outputvalues[j] > flickcounter)	//if the count is greater than the random nibble assigned 
							outputbufferA = (outputbufferA | (1<<j)); //to that LED, turn it off (1 is off).
					}
					if (outputvalues[9] > flickcounter)	//B is handles separately, because it's easier that way
						outputbufferB = 0x02;
					PORTA = (outputbufferA|Acandles);	
					PORTB = ((outputbufferB|Bcandles)&0x02);
					flickcounter--;
				}
				flickcounter = 0x0F;
			}
		}
				shutdown = 0; //it's time for the candles to burn down
				sleep_enable();
				sleep = 1;
				while (sleep) //stay in this loop until it's time to wake up (the pad is pressed).
				{
					PORTA = 0XFF;			//Turn off all LEDs.
					PORTB = PORTB | 0X02;
					DDRA = 0X00;
					DDRB = 0X00;
					day = 1;
					sleep_cpu();
					cli();
					if (touching())
					{
						changeday(1);
						display();
						sleep = 0;
						letgo(0);
					}
					slowcharge();
					sei();
					DDRB =  0xFF;	//All of port A is used, only pin 1 of port B is.
					PORTB = 0X00;	//note that 1 is off and 0 is on
					DDRA = 0XFF;
					PORTA = 0X00;
				}
				sleep_disable();
		 				


		
		
	}
		

    return 0;   /* never reached */
}

